"""
adamspy is a set of python tools for manipulating MSC Adams files and data

Available subpackages
---------------------
adripy
    A set of python tools for manipulating MSC Adams Drill files and data

Available modules
-----------------
adamspy
    Functions for manipulating adams files and data
"""
from . import adripy
from . import waterfall
name = 'adamspy'


